#/********************************************************************
# Cayenue/cayenue/panels/settings/storage.py 
#
# Copyright (c) 2024  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from PyQt6.QtWidgets import QMessageBox, QSpinBox, \
    QGridLayout, QWidget, QCheckBox, QLabel, QMessageBox, QGroupBox
from PyQt6.QtCore import QStandardPaths
from loguru import logger
from pathlib import Path
import shutil
from cayenue.components import DirectorySelector
from PyQt6.QtCore import pyqtSignal, QObject
import threading
import os
import tempfile

class StorageSignals(QObject):
    updateDiskUsage = pyqtSignal()
    showWaitDialog = pyqtSignal(str)

class StorageOptions(QWidget):
    def __init__(self, mw):
        super().__init__()
        self.mw = mw
        self.max_size = 100

        self.archiveKey = "settings/archive"
        self.pictureKey = "settings/picture"
        self.diskLimitKey = "settings/diskLimit"
        self.maxFileDurationKey = "settings/maxFileDuration"
        self.mangageDiskUsagekey = "settings/manageDiskUsage"
        self.writeBufferSizeKey = "setting/writeBufferSize"

        self.signals = StorageSignals()
        self.signals.updateDiskUsage.connect(self.updateDiskUsage)
        self.signals.showWaitDialog.connect(self.mw.dlgWait.show)

        video_dirs = QStandardPaths.standardLocations(QStandardPaths.StandardLocation.MoviesLocation)
        self.dirArchive = DirectorySelector(mw, self.archiveKey, "Archive Dir", video_dirs[0])
        self.dirArchive.signals.dirChanged.connect(self.dirArchiveChanged)

        picture_dirs = QStandardPaths.standardLocations(QStandardPaths.StandardLocation.PicturesLocation)
        self.dirPictures = DirectorySelector(mw, self.pictureKey, "Picture Dir", picture_dirs[0])
        self.dirPictures.signals.dirChanged.connect(self.dirPicturesChanged)

        self.grpDiskUsage = QGroupBox()

        self.chkManageDiskUsage = QCheckBox()
        self.chkManageDiskUsage.setChecked(bool(int(self.mw.settings.value(self.mangageDiskUsagekey, 0))))
        self.chkManageDiskUsage.clicked.connect(self.chkManageDiskUsageChanged)

        self.spnDiskLimit = QSpinBox()
        self.spnDiskLimit.valueChanged.connect(self.spnDiskLimitChanged)

        self.spnMaxFileDuration = QSpinBox()
        self.spnMaxFileDuration.setMaximum(60)
        self.spnMaxFileDuration.setValue(int(self.mw.settings.value(self.maxFileDurationKey, 15)))
        self.spnMaxFileDuration.valueChanged.connect(self.spnMaxFileDurationChanged)
        lblMaxFileDuration = QLabel("Max File Duration (minutes)")

        self.spnWriteBufferSize = QSpinBox()
        self.spnWriteBufferSize.setValue(int(self.mw.settings.value(self.writeBufferSizeKey, 10)))
        self.spnWriteBufferSize.valueChanged.connect(self.spnWriteBufferSizeChanged)
        lblWriteBuffferSize = QLabel("File Write Buffer Size (GB)")

        self.updateDiskUsage()
        #disk_limit = min(int(self.mw.settings.value(self.diskLimitKey, 100)), self.max_size)
        disk_limit = int(self.mw.settings.value(self.diskLimitKey, 100))
        self.spnDiskLimit.setValue(disk_limit)

        lytDiskUsage = QGridLayout(self.grpDiskUsage)
        lytDiskUsage.addWidget(self.chkManageDiskUsage, 0, 0, 1, 1)
        lytDiskUsage.addWidget(self.spnDiskLimit,       0, 2, 1, 1)
        lytDiskUsage.addWidget(QLabel("GB"),            0, 3, 1, 1)
        lytDiskUsage.addWidget(self.dirArchive,         1, 0, 1, 4)
        lytDiskUsage.addWidget(self.dirPictures,        2, 0, 1, 4)
        lytDiskUsage.addWidget(lblMaxFileDuration,      3, 0, 1, 2)
        lytDiskUsage.addWidget(self.spnMaxFileDuration, 3, 2, 1, 1)
        lytDiskUsage.addWidget(lblWriteBuffferSize,     4, 0, 1, 2)
        lytDiskUsage.addWidget(self.spnWriteBufferSize, 4, 2, 1, 1)
        lytDiskUsage.setColumnStretch(2, 10)

        lytMain = QGridLayout(self)
        lytMain.addWidget(self.grpDiskUsage, 0, 0, 1, 1)
        lytMain.addWidget(QLabel(),          1, 0, 1, 1)
        lytMain.addWidget(QLabel(),          2, 0, 1, 1)
        lytMain.setRowStretch(2, 10)


    def spnDiskLimitChanged(self, value):
        self.mw.settings.setValue(self.diskLimitKey, value)

    def spnMaxFileDurationChanged(self, value):
        self.mw.settings.setValue(self.maxFileDurationKey, value)
        self.mw.STD_FILE_DURATION = value * 60

    def chkManageDiskUsageChanged(self):
        if self.chkManageDiskUsage.isChecked():
            ret = QMessageBox.warning(self, "** WARNING **",
                                        "You are giving full control of both the video and picture archive directories to this program.  "
                                        "Any files contained within those directories or their sub-directories are subject to deletion.  "
                                        "You should only enable this feature if you are sure that this is ok.\n\n"
                                        "Are you sure you want to continue?",
                                        QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)
            if ret == QMessageBox.StandardButton.Cancel:
                self.chkManageDiskUsage.setChecked(False)
            else:
                self.signals.showWaitDialog.emit("Please wait for disk check to complete")
                thread = threading.Thread(target=self.calculateDiskUsage)
                thread.start()
                #self.updateDiskUsage()
        else:
            self.updateDiskUsage()
        self.mw.settings.setValue(self.mangageDiskUsagekey, int(self.chkManageDiskUsage.isChecked()))

    def dirArchiveChanged(self, path):
        logger.debug(f'Video archive directory changed to {path}')
        self.dirArchive.txtDirectory.setText(path)
        self.mw.settings.setValue(self.archiveKey, path)
        self.updateDiskUsage()
        #self.mw.filePanel.dirArchive.txtDirectory.setText(path)
        #self.mw.filePanel.dirChanged(path)

    def dirPicturesChanged(self, path):
        logger.debug(f'Picture directory changed to {path}')
        self.dirPictures.txtDirectory.setText(path)
        self.mw.settings.setValue(self.pictureKey, path)
        #self.mw.filePanel.control.dlgPicture.dirPictures.txtDirectory.setText(path)
        #self.mw.filePanel.control.dlgPicture.dirChanged(path)

    def spnWriteBufferSizeChanged(self, value):
        self.mw.settings.setValue(self.writeBufferSizeKey, value)
        self.updateDiskUsage()

    def calculateDiskUsage(self):
        if not self.mw.settings_profile == "gui":
            return
        dir = self.dirArchive.txtDirectory.text()
        size = sum(f.stat().st_size for f in Path(dir).rglob('*') if f.is_file())
        _, _, free = shutil.disk_usage(dir)
        buffer_size = self.spnWriteBufferSize.value() * 1_000_000_000
        self.max_size = int((free + size - buffer_size)/1_000_000_000)
        self.lblManage = f'Auto Manage (max {self.max_size} GB)'
        self.chkManageDiskUsage.setText(self.lblManage)
        self.spnDiskLimit.setMaximum(self.max_size)
        if self.spnDiskLimit.value() > self.max_size:
            self.spnDiskLimit.setValue(self.max_size)
        str_size_in_GB = "{:.2f}".format(size / 1_000_000_000)
        self.grpDiskUsage.setTitle(f'Disk Usage (currently {str_size_in_GB} GB)')
        self.mw.signals.hideWaitDialog.emit()

    def updateDiskUsage(self):
        if self.chkManageDiskUsage.isChecked():
            self.calculateDiskUsage()
        else:
            self.spnDiskLimit.setMaximum(self.max_size)
            self.grpDiskUsage.setTitle("Disk Usage Auto Manage")
            self.chkManageDiskUsage.setText("Auto Manage")


from time import sleep
from loguru import logger
import libonvif as onvif
from pathlib import Path
import os
import subprocess
from datetime import datetime
from cayenue.enums import ProxyType, SnapshotAuth
import platform
import webbrowser
import requests
from requests.auth import HTTPDigestAuth
from urllib.parse import urlparse, parse_qs
import threading


class Snapshot():
    def __init__(self, mw):
        self.mw = mw

    def getSnapshot(self, profile, filename, auth_type):
        parsed = urlparse(profile.snapshot_uri())

        if auth_type == SnapshotAuth.NOT_WORKING:
            return False

        if auth_type == SnapshotAuth.UNKNOWN:
            if self.getSnapshot(profile, filename, SnapshotAuth.BASIC):
                camera = self.mw.cameraPanel.getCamera(profile.uri())
                camera.setSnapshotAuth(SnapshotAuth.BASIC)
                logger.debug(f"camera {camera.name()} snapshot auth type set to BASIC")
                return True

            else:
                if self.getSnapshot(profile, filename, SnapshotAuth.DIGEST):
                    camera = self.mw.cameraPanel.getCamera(profile.uri())
                    camera.setSnapshotAuth(SnapshotAuth.DIGEST)
                    logger.debug(f"camera {camera.name()} snapshot auth type set to DIGEST")
                    return True

        if auth_type == SnapshotAuth.BASIC:
            try:
                simple_url = f"{parsed.scheme}://{profile.username()}:{profile.password()}@{parsed.netloc}{parsed.path}?{parsed.query}"
                response = requests.get(simple_url, timeout=5)
                if response.status_code == 200:
                    with open(filename, 'wb') as file:
                        file.write(response.content)
                    #logger.debug(f"Image downloaded successfully as {filename}")
                    return True
                else:
                    logger.debug(f"get snapshot BASIC auth status code: {response}")
                    return False
            except Exception as ex:
                logger.debug(f'set snapshot BASIC auth exception {ex}')
            return False

        if auth_type == SnapshotAuth.DIGEST:
            try:
                base_url = f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
                params = {k: v[0] for k, v in parse_qs(parsed.query).items()}
                response = requests.get(base_url, params=params, auth=HTTPDigestAuth(profile.username(), profile.password()), timeout=5)
                if response.status_code == 200:
                    with open(filename, 'wb') as file:
                        file.write(response.content)
                    #logger.debug(f"Image downloaded successfully as {filename}")
                    return True
                else:
                    logger.debug(f"get snapshot DIGEST auth status code: {response}")
                    return False
            except Exception as ex:
                logger.debug(f'set snapshot DIGEST auth exception {ex}')
            return False

        return False

    def getSnapshotBuffered(self, profile, buffer, auth_type):
        print("Here we are at getSnapshotBuffered", auth_type)
        parsed = urlparse(profile.snapshot_uri())

        if auth_type == SnapshotAuth.NOT_WORKING:
            return False

        if auth_type == SnapshotAuth.UNKNOWN:
            if self.getSnapshotBuffered(profile, buffer, SnapshotAuth.BASIC):
                #camera = self.mw.cameraPanel.getCamera(profile.uri())
                #camera.setSnapshotAuth(SnapshotAuth.BASIC)
                #logger.debug(f"camera {camera.name()} snapshot auth type set to BASIC")
                return True

            else:
                if self.getSnapshotBuffered(profile, buffer, SnapshotAuth.DIGEST):
                    #camera = self.mw.cameraPanel.getCamera(profile.uri())
                    #camera.setSnapshotAuth(SnapshotAuth.DIGEST)
                    #logger.debug(f"camera {camera.name()} snapshot auth type set to DIGEST")
                    return True

        if auth_type == SnapshotAuth.BASIC:
            try:
                simple_url = f"{parsed.scheme}://{profile.username()}:{profile.password()}@{parsed.netloc}{parsed.path}?{parsed.query}"
                response = requests.get(simple_url, timeout=5)
                if response.status_code == 200:
                    
                    print("Writing buffer from BASIC AUTH")
                    buffer.write(response.content)
                    
                    #logger.debug(f"Image downloaded successfully as {filename}")
                    return True
                else:
                    logger.debug(f"get snapshot BASIC auth status code: {response}")
                    return False
            except Exception as ex:
                logger.debug(f'set snapshot BASIC auth exception {ex}')
            return False

        if auth_type == SnapshotAuth.DIGEST:
            try:
                base_url = f"{parsed.scheme}://{parsed.netloc}{parsed.path}"
                print("BASE URL:", base_url)
                params = {k: v[0] for k, v in parse_qs(parsed.query).items()}
                response = requests.get(base_url, params=params, auth=HTTPDigestAuth(profile.username(), profile.password()), timeout=5)
                if response.status_code == 200:
                    
                    print("Writing buffer from DIGEST AUTH")
                    buffer.write(response.content)
                    #with open("/home/stephen/Pictures/sample.jpg", "wb") as file:
                        #file.write(response.content)
                        #file.write(buffer.getvalue())
                    #print("where is the file")

                    #logger.debug(f"Image downloaded successfully as {filename}")
                    return True
                else:
                    logger.debug(f"get snapshot DIGEST auth status code: {response}")
                    return False
            except Exception as ex:
                logger.debug(f'set snapshot DIGEST auth exception {ex}')
            return False

        return False

    def getBufferedSnapshot(self, profile, buffer, camera):
        camera.setSnapshotAuth(SnapshotAuth.UNKNOWN)
        print("call came into getBufferedSnapshot", camera.name(), camera.snapshotAuth)
        if not self.getSnapshotBuffered(profile, buffer, camera.snapshotAuth):
            if camera.snapshotAuth is not SnapshotAuth.NOT_WORKING:
                logger.debug(f'camera {camera.name()} snapshot auth type set to NOT_WORKING')
                camera.setSnapshotAuth(SnapshotAuth.NOT_WORKING)
            #player.image.save(filename)
            #return buffer


    def __call__(self, profile, filename, camera, player):
        if not camera.systemTabSettings.remote_snapshot:
            player.image.save(filename)
            return
        if self.mw.settingsPanel.proxy.proxyType == ProxyType.CLIENT:
            print("CLIENT PROXY", camera.name())
            profile.setUserData(filename)
            self.mw.client.transmit(bytearray(f"SNAPSHOT\n\n{profile.toJSON()}\r\n", 'utf-8'))
            return
        if not self.getSnapshot(profile, filename, camera.snapshotAuth):
            if camera.snapshotAuth is not SnapshotAuth.NOT_WORKING:
                logger.debug(f'camera {camera.name()} snapshot auth type set to NOT_WORKING')
                camera.setSnapshotAuth(SnapshotAuth.NOT_WORKING)
            player.image.save(filename)
            #logger.debug(f'Raw stream image saved as {filename}')



#/********************************************************************
# Cayenue/cayenue/panels/camera/camerapanel.py 
#
# Copyright (c) 2023  Stephen Rhodes
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#*********************************************************************/

from time import sleep
from PyQt6.QtWidgets import QPushButton, QGridLayout, QWidget, QSlider, \
    QListWidget, QTabWidget, QMessageBox, QMenu, QFileDialog
from PyQt6.QtGui import QAction
from PyQt6.QtCore import Qt, pyqtSignal, QObject, QTimer, QSettings
from . import NetworkTab, ImageTab, VideoTab, PTZTab, SystemTab, LoginDialog, \
    Session, Camera
from loguru import logger
import libonvif as onvif
from pathlib import Path
import os
import subprocess
from datetime import datetime
from cayenue.enums import ProxyType, SnapshotAuth
import platform
import webbrowser
import requests
from requests.auth import HTTPDigestAuth
from urllib.parse import urlparse, parse_qs
import threading

class CameraList(QListWidget):
    def __init__(self, mw):
        super().__init__()
        #self.signals = CameraPanelSignals()
        self.setSortingEnabled(True)
        self.mw = mw

    def focusInEvent(self, event):
        if self.currentRow() == -1:
            self.setCurrentRow(0)
        super().focusInEvent(event)

    def keyPressEvent(self, event):
        match event.key():
            case Qt.Key.Key_Return:
                if camera := self.currentItem():
                    if not camera.editing():
                        self.itemDoubleClicked.emit(camera)
                        self.setFocus()

            case Qt.Key.Key_Escape:
                if self.mw.isFullScreen():
                    self.mw.showNormal()
                elif self.mw.focus_window and self.mw.focus_window.isVisible():
                    self.mw.focus_window.close()
                elif camera := self.currentItem():
                    players = self.mw.pm.getStreamPairPlayers(camera.uri())
                    if len(players):
                        self.mw.cameraPanel.btnStopClicked()

            case Qt.Key.Key_Delete:
                self.remove()

            case Qt.Key.Key_F2:
                self.rename()

            case Qt.Key.Key_F1:
                self.info()

            case Qt.Key.Key_A:
                if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
                    self.mw.cameraPanel.btnStopAllClicked()
            case Qt.Key.Key_F:
                if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
                    self.mw.cameraPanel.btnHistoryClicked()
            case Qt.Key.Key_S:
                if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
                    self.mw.cameraPanel.btnSnapshotClicked()
            case Qt.Key.Key_R:
                if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
                    self.mw.cameraPanel.btnRecordClicked()

        return super().keyPressEvent(event)
    
    def remove(self):
        if camera := self.currentItem():
            if self.mw.pm.getPlayer(camera.uri()):
                ret = QMessageBox.warning(self, camera.name(),
                                            "Camera is currently playing. Please stop before deleting.",
                                            QMessageBox.StandardButton.Ok)

                return
            else:
                ret = QMessageBox.warning(self, camera.name(),
                                            "Removing this camera from the list.\n"
                                            "Are you sure you want to continue?",
                                            QMessageBox.StandardButton.Ok | QMessageBox.StandardButton.Cancel)

                if ret != QMessageBox.StandardButton.Ok:
                    return

            row = self.currentRow()
            if row > -1:
                if camera.filled:
                    camera = self.takeItem(row)
                    for profile in camera.profiles:
                        self.mw.proxies.pop(profile.stream_uri(), None)

                    if self.mw.settingsPanel.proxy.proxyType == ProxyType.SERVER:
                        self.mw.settingsPanel.proxy.setMediaMTXProxies()

                else:
                    ret = QMessageBox.warning(self, camera.name(),
                                                "The program is currently communicating with the camera. Please wait before deleting.",
                                                QMessageBox.StandardButton.Ok)

        if not self.count():
            data = onvif.Data()
            self.mw.cameraPanel.signals.fill.emit(data)

        self.mw.cameraPanel.saveCameraList()

    def info(self):
        msg = ""
        if camera := self.currentItem():
            players = self.mw.pm.getStreamPairPlayers(camera.uri())
            if not len(players):
                msg = "Start camera to get stream info"
            for i, player in enumerate(players):
                if i == 0:
                    msg += "<h2>Display Stream</h2>"
                    msg += player.getStreamInfo()
                    msg += "\n"
                if i == 1:
                    msg += "<h2>Record Stream</h2>"
                    msg += player.getStreamInfo()
                    msg += "\n"
        msgBox = QMessageBox(self)
        msgBox.setWindowTitle("Stream Info")
        msgBox.setText(msg)
        msgBox.setTextFormat(Qt.TextFormat.RichText)
        msgBox.exec()
    
    def rename(self):
        if camera := self.currentItem():
            camera.setFlags(camera.flags() | Qt.ItemFlag.ItemIsEditable)
            index = self.currentIndex()
            if index.isValid():
                self.edit(index)

    def password(self):
        if camera := self.currentItem():
            self.mw.settings.setValue(f'{camera.xaddrs()}/alternateUsername', camera.onvif_data.username())
            self.mw.settings.setValue(f'{camera.xaddrs()}/alternatePassword', camera.onvif_data.password())
            logger.debug(f'Alternate password set for camera {camera.name()}')

    def closeEditor(self, editor, hint):
        if camera := self.currentItem():
            camera.onvif_data.alias = editor.text()
            self.mw.settings.setValue(f'{camera.serial_number()}/Alias', editor.text())
            camera.setFlags(camera.flags() & ~Qt.ItemFlag.ItemIsEditable)
        return super().closeEditor(editor, hint)
    
    def startCamera(self):
        if camera := self.currentItem():
            self.mw.cameraPanel.onItemDoubleClicked(camera)

    def stopCamera(self):
        if camera := self.currentItem():
            self.mw.cameraPanel.onItemDoubleClicked(camera)

